'' =================================================================================================
''
''   File....... hc-8+_sio_template.spin
''   Purpose.... Allows custom HC-8+ to control AP-16+ and remote HC-8+s
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2013 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 30 JAN 2013
''
'' =================================================================================================


con

  PCB_REV = "C"


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  

con

  RX1     = 31                                                  ' programming / debug port
  TX1     = 30
  
  SDA     = 29                                                  ' eeprom
  SCL     = 28

  OUT_EN  = 27                                                  ' outputs enable pin (low)     
  
  SIO     = 26                                                  ' TTL serial io
  
  RX2     = 23                                                  ' RS-485 port
  RX2_EN  = 22
  TX2_EN  = 21
  TX2     = 20

  R_LED   = 18                                                  ' bi-color LED
  G_LED   = 17

  OUT7    = 15                                                  ' outputs
  OUT6    = 14   
  OUT5    = 13 
  OUT4    = 12
  OUT3    = 11
  OUT2    = 10   
  OUT1    =  9 
  OUT0    =  8
          
  OPT_BR  =  7                                                  ' options swithes
  OPT_A1  =  6   
  OPT_A0  =  5
  OPT_SM  =  4

  DMX_A8  =  3                                                  ' high-bit off DMX address

  DO_165  =  2                                                  ' control lines for 74x165
  CLK_165 =  1 
  LD_165  =  0


  ' for EFX-TEK GameOn adapter module (#99001)

  WII_DAT =  OPT_A0
  WII_CLK =  OPT_SM 

  
  ' for EFX-TEK uSD adapter module (#99002)
  ' -- must open all option switches when installed

  SD_CS   =  OPT_BR                                       
  SD_DI   =  OPT_A1  
  SD_CLK  =  OPT_A0  
  SD_DO   =  OPT_SM


con

   #0, OFF, RED, GRN, YEL                                       ' for r/g led

   #0, NOCON, RX, TX, RXTX                                      ' for rs-485

   #0, LSBFIRST, MSBFIRST                                       ' for 74x165 shift-register                                               


con

   #1, HOME, GOTOXY, #8, BKSP, TAB, LF, CLREOL, CLRDN, CR       ' PST formmatting control
  #14, GOTOX, GOTOY, CLS


obj

  term  : "fullduplexserial64"                                  ' terminal IO   
  com1  : "fullduplexserial64"                                  ' half-duplex serial IO
  rr    : "realrandom"                                          ' real random from hardware
  prng  : "efx_prng"                                            ' pseudo-random number generator
  outs  : "efx_pwm8"                                            ' pwm driver for outputs
    

var

  byte  brightness[8]                                           ' channel brightness


dat

AP16                    byte    "!AP16", 0
HC8                     byte    "!HC8", 0

Version                 byte    "???", 0


pub main | t

  setup                                                         ' configure hardware


  t := cnt
  repeat

    ' code here

    waitcnt(t += (100 * MS_001))                                ' update every 100ms


pub set_channel(ch, state) | addr

'' Sets output channel state (digital)
'' -- works like high and low across multiple boards
'' -- allows up to 40 "channels" (local + 4 remote HC-8+s)
'' -- ch is 0 to 39
'' -- state is 0 or 1 (non-zero converted to 1)

  if ((ch => 0) and (ch =< 7))                                  ' local output?
    if (state)           
      outs.high(ch)   
    else                 
      outs.low(ch)

  elseif ((ch => 8) and (ch =< 39))                             ' remote output?
    addr := (ch >> 3) - 1                                       ' calculate board address  
    ch  &= %111                                                 ' convert to remote ch # 
    if (state)                                                  ' make digital
      state := 1

    hc8_set_digital(addr, ch, state)  


pub set_level(ch, level) | addr

'' Sets output channel level (analog)
'' -- allows up to 40 "channels" (local + 4 remote HC-8+s)
'' -- ch is 0 to 39
'' -- level is 0 to 255 (off to full on)

  level := 0 #> level <# 255                                    ' force to legal range

  if ((ch => 0) and (ch =< 7))                                  ' local output?
    outs.set(ch, level)

  elseif ((ch => 8) and (ch =< 39))                             ' remote output?
    addr := (ch >> 3) - 1                                       ' calculate board address  
    ch  &= %111                                                 ' convert to remote ch # 
    hc8_set_level(addr, ch, level)     

  
pub setup

'' Setup HC-8+
'' -- configures local IO, randomizer, serial ports

  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode                                  
  low(OUT_EN)                                                   ' enable 74x245 

  high(LD_165)                                                  ' setup x165 io pins
  low(CLK_165)
  input(DO_165)

  set_rgled(OFF)
  
  set_rs485(NOCON)                                              ' RS-485 not connected

  term.start(RX1, TX1, %0000, 115_200)                          ' terminal serial         
  com1.start(SIO, SIO, %1100, 38_400)                           ' half-duplex serial 

  rr.start                                                      ' start hardware randomizer
  prng.seed($EF8_7E3, rr.random, rr.random, rr.random, 0)       ' seed software random
  rr.stop                                                       ' unload hardware randomizer

  outs.start(8, OUT0)                                           ' start local outputs driver

  pause(10)                                                     ' let serial lines stabilize
  term.rxflush                                                  ' clear rx buffers
  com1.rxflush

  
con

  { ------------------------------- }
  {                                 }
  {  A P - 1 6 +   R O U T I N E S  } 
  {                                 }
  { ------------------------------- }



pub ap16_reset(addr)

'' Stops and resets AP-16+ at addr

  com1.str(@AP16)
  com1.tx(addr)
  com1.tx("X")
  
  
pub ap16_version(addr) | idx, c

'' Updates version string from AP-16+ at addr
'' -- returns 0 if good response
'' -- returns -1 if no response

  com1.str(@AP16)
  com1.tx(addr)
  com1.tx("V")
  com1.rxflush

  repeat idx from 0 to 2                                        ' expecting 3 characters
    c := com1.rxtime(25)                                        ' allow up to 25ms response
    if (c => 0)                                                 ' if good response
      byte[@Version][idx] := c                                  '  add ch to version string             
    else                                                        ' else
      bytefill(@Version, string("???"), 3)                      '  clear version string
      return -1                                                 '  report timeout error

  return 0


pub ap16_status(addr)

'' Returns AP-16+ status byte

  com1.str(@AP16)
  com1.tx(addr)
  com1.tx("G")
  com1.rxflush

  return com1.rxtime(25)


pub ap16_playing(addr) | status

'' Returns playing status from AP-16+ at addr

  return ((ap16_status(addr) & %1000_0000) > 0)
    
    
pub ap16_sfx(addr, sfx, rpts)

'' Plays SFX file on AP-16+ at addr

  com1.str(@AP16)
  com1.tx(addr)
  com1.str(string("PS"))
  com1.tx(sfx)
  com1.tx(rpts)   


pub ap16_aux(addr, aux, rpts)

'' Plays AUX file on AP-16+ at addr

  com1.str(@AP16)
  com1.tx(addr)
  com1.str(string("PA"))
  com1.tx(aux)
  com1.tx(rpts)


pub ap16_wav(addr, spntr, rpts)

'' Plays WAV file (name at spntr) on AP-16+ at addr

  com1.str(@AP16)
  com1.tx(addr)
  com1.str(string("PW"))
  com1.str(spntr)
  com1.tx(13)      
  com1.tx(rpts)


con

  { ----------------------------- }
  {                               }
  {  H C - 8 +   R O U T I N E S  } 
  {                               }
  { ----------------------------- }


pub hc8_reset(addr)

'' Resets HC-8+ at addr

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("X")


pub hc8_version(addr) | idx, c

'' Updates version string from HC-8+ at addr
'' -- returns 0 if good response
'' -- returns -1 if no response

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("V")
  com1.rxflush

  repeat idx from 0 to 2                    
    c := com1.rxtime(25)                    
    if (c => 0)                             
      byte[@Version][idx] := c                         
    else                                    
      bytefill(@Version, string("???"), 3)  
      return -1                             

  return 0

    
pub hc8_set_level(addr, ch, level)     

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("L")
  com1.tx(ch)
  com1.tx(level)


pub hc8_set_ilevels(addr) | idx

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("L")
  com1.tx("I")
  repeat idx from 0 to 7
    com1.tx(brightness[idx])
  

pub hc8_randomize(addr, mask)

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("L")
  com1.tx("R")
  com1.tx(mask)


pub hc8_set_digital(addr, ch, state)

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("D")
  com1.tx(ch)
  com1.tx(state)


pub hc8_digital_select(addr, states, mask)

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("D")
  com1.tx("S")  
  com1.tx(states)
  com1.tx(mask)


pub hc8_digital_random(addr, mask)

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("D")
  com1.tx("R")  
  com1.tx(mask)


pub hc8_fade0(addr, ch, levela, levelb, seconds)

' low-resolution fade

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("F")
  com1.tx("L")
  com1.tx(ch)
  com1.tx(levela)
  com1.tx(levelb)
  com1.tx(seconds)
  

pub hc8_fade1(addr, ch, levela, levelb, tenths)

' medium-resolution fade

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("F")
  com1.tx("M")
  com1.tx(ch)
  com1.tx(levela)
  com1.tx(levelb)
  com1.tx(tenths)


pub hc8_fade2(addr, ch, levela, levelb, ms)

' high-resolution fade

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("F")
  com1.tx("L")
  com1.tx(ch)
  com1.tx(levela)
  com1.tx(levelb)
  com1.tx(ms.byte[0])
  com1.tx(ms.byte[1])


pub hc8_cross_fade(addr, cha, chb, tenths)

' cross-fade (medium resolution timing)

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("F")
  com1.tx("X")
  com1.tx(cha)
  com1.tx(chb)
  com1.tx(tenths)


pub hc8_get_channel(addr, ch) | lvl

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("G")

  com1.rxflush
  return com1.rxtime(25)


pub hc8_read_ttl(addr) | ttl

  com1.str(@HC8)
  com1.tx(addr)
  com1.tx("I")
  
  com1.rxflush
  return com1.rxtime(25)


con

  { --------------------------------- }
  {                                   }
  {  S U P P O R T   R O U T I N E S  } 
  {                                   }
  { --------------------------------- }


var

  long  lastscan                                                ' ttl/DMX input scan


pub pause(ms) | t

'' Delay program in milliseconds

  if (ms < 1)                                                   ' delay must be at least 1
    return
  else
    t := cnt - 1792                                             ' sync with system counter
    repeat ms                                                   ' run delay
      waitcnt(t += MS_001)


pub high(pin)

'' Makes pin output and high

  outa[pin] := 1
  dira[pin] := 1


pub low(pin)

'' Makes pin output and low

  outa[pin] := 0
  dira[pin] := 1


pub toggle(pin)

'' Toggles output pin

  !outa[pin]                                                    ' invert state
  dira[pin] := 1  
  

pub input(pin)

  dira[pin] := 0

  return ina[pin]


pub ttl_pin | btns

'' Reads TTL header for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  btns := ttl_inputs(true)                                      ' read ttl inputs

  if (bit_count(btns) == 1)                                     ' only one button pressed?
    return bit_pos(btns, 0)                                     '  yes, return button #
  else
    return -1                                                   '  none or bad input


pub ttl_inputs(rescan) 

'' Return TTL inputs status
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss
   
  return (lastscan & $FF)                                       ' return TTL input bits


pub dmx_address(rescan) 

'' Return DMX address switch setting
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss

  return (lastscan >> 8) & $01FF                                ' return DMX address bits
 

pub x165_in(bits, mode) | tmp165

'' Input value from 74x165(s)      
                                                                                          
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                                                         
                                                                                          
  tmp165 := 0                                                   ' clear workspace
  bits := 1 #> bits <# 32                                       ' limit to legal value         
  repeat bits                                                   ' read n bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                       ' get new bit             
    outa[CLK_165] := 1                                          ' blip clock              
    outa[CLK_165] := 0  

  if (mode == LSBFIRST)                                         ' LSBFIRST result?
    tmp165 ><= bits                                             ' reverse bits
   
  return tmp165
  

pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                 ' force to input mode

  return !ina[OPT_BR..OPT_SM]                                   ' return state (1 = on)


pub set_rgled(state)

'' Sets R/G LED
'' -- direct control only
'' -- for red/green/yellow use "efx_rgled" driver and methods

  case state
    OFF : outa[R_LED..G_LED] := %00
    RED : outa[R_LED..G_LED] := %10
    GRN : outa[R_LED..G_LED] := %01
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires PASM driver

  dira[R_LED..G_LED] := %11
      

pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(value) | bc

'' Returns # of bits set (1) in value

  bc := 0                                                       ' clear count
  repeat 32                                                     ' test all bits
    bc += (value & %1)                                          ' add bit value to count
    value >>= 1                                                 ' next bit

  return bc


pub bit_pos(value, mode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb 
'' -- -1 = no bits set

  if (value == 0)                                               ' if no bits                                   
    return -1                                                   '  return -1

  else
    if (mode == 0)                                              ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)
    else
      return (>|value - 1)


pub bit_val(value, pos)

'' Returns bit (0..1) from pos (0..31) in value 

  if ((pos => 0) and (pos =< 31))
    return (value >> pos) & 1
  else
    return 0

    
dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      